import { h } from '../anjuna-charts.core.js';

import { a as d3 } from './chunk-613005d7.js';
import { a as ChartDataService, b as SeriesStyle } from './chunk-86ee18a4.js';

var SeriesType;
(function (SeriesType) {
    SeriesType[SeriesType["Line"] = 0] = "Line";
    SeriesType[SeriesType["Column"] = 1] = "Column";
    SeriesType[SeriesType["Stacked"] = 2] = "Stacked";
})(SeriesType || (SeriesType = {}));

class ChartD3Service {
    static addArea(svg, line, y, data, color) {
        const area = d3.area()
            .curve(d3.curveMonotoneX)
            .defined(line.defined())
            .x(line.x())
            .y1(line.y())
            .y0(y(0));
        svg.append('path')
            .data([data])
            .attr('class', 'series area')
            .attr('fill', color)
            .attr('d', area);
        return area;
    }
    static addStackedColumn(svg, x, y, data, color) {
        const seriesContainers = svg
            .append('g')
            .attr('class', 'series-container')
            .selectAll('.series-container')
            .data([data])
            .enter().append('g')
            .attr('class', 'group')
            .style('fill', color);
        const chartBars = seriesContainers
            .selectAll('rect')
            .data((d) => d)
            .enter().append('rect');
        chartBars
            .attr('class', (d, i) => {
            return d[0] < 0 ? `bar negative bar-${i}` : `bar positive bar-${i}`;
        })
            .attr('width', x.bandwidth() - ChartD3Service.groupPadding)
            .attr('x', (d) => x(d.data.categoryValue) + (ChartD3Service.groupPadding / 2))
            .attr('y', y(0))
            .attr('height', 0);
        chartBars
            .transition()
            .duration(ChartD3Service.animationLength)
            .attr('y', (d) => y(d[1]))
            .attr('height', (d) => y(d[0]) - y(d[1]));
        return chartBars;
    }
    static addColumn(svg, x, x1, y, data, color, _dims, index) {
        const seriesContainers = svg
            .append('g')
            .attr('class', 'series-container')
            .selectAll('.series-container')
            .data([data])
            .enter().append('g');
        const chartBars = seriesContainers
            .attr('class', 'series')
            .attr('transform', 'translate(' + x1(index) + ',0)')
            .style('fill', color)
            .selectAll('rect')
            .data((d) => d)
            .enter().append('rect');
        chartBars
            .attr('class', (d) => {
            return d.y < 0 ? 'bar negative' : 'bar positive';
        })
            .style('opacity', (d) => {
            return d.y < 0 ? .6 : 1;
        })
            .attr('width', x1.bandwidth() - ChartD3Service.groupPadding)
            .attr('x', (d) => x(d.x) + (ChartD3Service.groupPadding / 2))
            .attr('y', () => y(0))
            .attr('height', 0);
        chartBars
            .transition()
            .duration(ChartD3Service.animationLength)
            .attr('y', (d) => d.y >= 0 ? (y(d.y)) : y(0))
            .attr('height', (d) => {
            return Math.max(Math.abs(y(d.y) - y(0)), 1);
        });
        return chartBars;
    }
    static addLine(svg, x, y, data, color) {
        let xOffset = 0;
        if (x.bandwidth) {
            xOffset = ((x.bandwidth()) / 2);
        }
        const valueLine = d3.line()
            .curve(d3.curveMonotoneX)
            .defined((d) => d.y !== null)
            .x((d) => x(d.x) + xOffset)
            .y((d) => y(d.y));
        svg.append('path')
            .data([data])
            .attr('class', 'line')
            .attr('stroke', color)
            .attr('d', valueLine);
        return valueLine;
    }
    static addOverlay(svg, dims) {
        const overlay = svg.append('rect')
            .attr('class', 'overlay')
            .attr('width', dims.width)
            .attr('height', dims.height);
        return overlay;
    }
    static addPoints(svg, x, y, data, color, index) {
        let xOffset = 0;
        if (x.bandwidth) {
            xOffset = ((x.bandwidth()) / 2);
        }
        svg.selectAll(`.point.series${index}`)
            .data(data)
            .enter()
            .append('svg:circle')
            .style('opacity', 0)
            .style('stroke', color)
            .attr('class', (_d, i) => `point point-${i}`)
            .attr('r', () => 1)
            .attr('cx', (d) => x(d.x) + xOffset)
            .attr('cy', (d) => y(d.y));
    }
    static addRevealAnimation(svg, dims, axisConfigs) {
        const delay = ChartD3Service.focusLength;
        const duration = ChartD3Service.animationLength;
        let curtainHeight = dims.height;
        let curtainWidth = dims.width;
        if (!axisConfigs.x.showAxis) {
            curtainHeight = curtainHeight + 1;
        }
        if (axisConfigs.y.showAxis) {
            curtainWidth = curtainWidth - (this.yPad * 2.5);
        }
        const curtain = svg.append('rect')
            .attr('x', -1 * dims.width)
            .attr('y', -1 * curtainHeight)
            .attr('height', curtainHeight)
            .attr('width', curtainWidth)
            .attr('class', 'curtain')
            .attr('transform', 'rotate(180)')
            .style('fill', '#ffffff');
        const t = svg.transition()
            .delay(delay)
            .duration(duration);
        t.select('rect.curtain')
            .attr('width', 0);
        t.select('line.guide')
            .attr('transform', 'translate(' + dims.width + ', 0)');
        return curtain;
    }
    static addStackedScrubbers(svg, x, data, dims, index) {
        const xOffset = ((x.bandwidth() + ChartD3Service.calculatePadding(data, dims)) / 2);
        svg.selectAll(`.scrubber.series${index}`)
            .data(data)
            .enter()
            .append('line')
            .attr('x1', (d) => x(d.data.categoryValue) + xOffset)
            .attr('y1', 0)
            .attr('x2', (d) => x(d.data.categoryValue) + xOffset)
            .attr('y2', dims.height)
            .attr('class', (_d, i) => `scrubber scrubber-${i}`)
            .attr('stroke', '#ffffff')
            .style('opacity', '0');
    }
    static addScrubbers(svg, x, data, dims, index) {
        svg.selectAll(`.scrubber.series${index}`)
            .data(data)
            .enter()
            .append('line')
            .attr('x1', (d) => x(d.x))
            .attr('y1', 0)
            .attr('x2', (d) => x(d.x))
            .attr('y2', dims.height)
            .attr('class', (_d, i) => `scrubber scrubber-${i}`);
    }
    static getX1(x, data, multiSeries) {
        const seriesCount = (multiSeries) ? data.length : 1;
        return d3.scaleBand()
            .rangeRound([0, x.bandwidth()])
            .domain(d3.range(seriesCount))
            .padding(0);
    }
    static addXAxis(svg, dims, data, axisConfigs, _minMax, primarySeries, y) {
        const x = ChartD3Service.axisTyper(primarySeries, data, dims)
            .range([this.calculateLeftPad(dims, axisConfigs, data), dims.width]);
        const uniqueX = Array.from(new Set(data.map(item => item.x.toString())));
        const availableSlots = (dims.width / 100);
        const tickModulo = Math.max(Math.floor(uniqueX.length / availableSlots), 1);
        const xAxis = (primarySeries === SeriesType.Line)
            ? d3.axisBottom(x)
            : d3.axisBottom(x).tickValues(x.domain().filter((_d, i) => !(i % tickModulo)));
        xAxis.tickSizeOuter(0)
            .tickFormat(ChartDataService.xFormatter.bind(this, ChartDataService.calculateInterval(data), axisConfigs.x))
            .scale(x);
        svg.append('g')
            .attr('class', 'x axis')
            .attr('transform', `translate(0,${dims.height})`)
            .call(xAxis);
        svg.select('.x.axis path')
            .attr('transform', 'translate(0,' + (y(0) - dims.height) + ')');
        if (!axisConfigs.x.showAxis) {
            svg.select('.x.axis').style('display', 'none');
        }
        if (!axisConfigs.x.showTicks) {
            svg.selectAll('.x.axis .tick').style('display', 'none');
        }
        if (primarySeries === SeriesType.Line) {
            const ticksLength = svg.selectAll('.x.axis .tick').size();
            const slots = (dims.width / 100);
            const modulo = Math.max(Math.floor(ticksLength / slots), 1);
            svg.selectAll('.x.axis .tick').filter((_d, i) => (i % modulo)).style('display', 'none');
        }
        return x;
    }
    static axisTyper(type, data, dims) {
        switch (type) {
            case SeriesType.Line:
                return d3.scaleTime()
                    .domain(ChartDataService.determineDateRange(data));
            default:
                const domain = data.map((d) => d.x.toString());
                return d3.scaleBand()
                    .padding(ChartD3Service.calculatePadding(data, dims))
                    .domain(domain);
        }
    }
    static calculatePadding(data, dims) {
        const colWidth = dims.width / (data.length - 1);
        return colWidth > 30 ? .05 : 0;
    }
    static addYAxis(svg, dims, data, axisConfigs, minMax) {
        const domain = (minMax) ? [minMax.min, minMax.max] : [Math.min(d3.min(data, (d) => d.y), 0), d3.max(data, (d) => d.y)];
        const y = d3.scaleLinear()
            .domain(domain)
            .nice()
            .range([dims.height, dims.margin.top]);
        const yAxis = d3.axisLeft(y)
            .tickSize(0)
            .ticks(this.calculateTicks(dims.height))
            .tickFormat(ChartDataService.yFormatter.bind(this, axisConfigs.y))
            .scale(y);
        svg.append('g')
            .attr('class', 'y axis')
            .attr('transform', `translate(${this.calculateLeftPad(dims, axisConfigs, data)},0)`)
            .call(yAxis);
        if (!axisConfigs.y.showAxis) {
            svg.select('.y.axis').style('display', 'none');
        }
        if (!axisConfigs.y.showTicks) {
            svg.selectAll('.y.axis .tick').style('display', 'none');
        }
        return y;
    }
    static createSVG(dims) {
        const svg = d3.create('svg');
        svg
            .attr('width', dims.width + dims.margin.left)
            .attr('height', dims.height + dims.margin.top)
            .attr('transform', `translate(${dims.margin.left}, ${dims.margin.top})`)
            .append('g');
        return svg;
    }
    static getIndexOfClosestPoint(x, data) {
        const x0 = x.invert(d3.mouse(d3.event.currentTarget)[0]);
        const bisector = d3.bisector((d) => d.x).left;
        const index = bisector(data, x0, 1);
        const d0 = data[index - 1];
        const d1 = data[index];
        const rangeValueOfFirst = x((d0) ? d0.x : null), rangeValueOfSecond = x((d1) ? d1.x : null), rangeValueOfMousePos = d3.mouse(d3.event.currentTarget)[0], closestIndex = Math.abs(rangeValueOfMousePos - rangeValueOfFirst) > Math.abs(rangeValueOfMousePos - rangeValueOfSecond) ? index : index - 1;
        return closestIndex;
    }
    static hideFocus(svg) {
        svg.selectAll('.point')
            .transition()
            .duration(ChartD3Service.focusLength)
            .attr('r', 1)
            .style('opacity', '0');
        svg.selectAll(`.scrubber`)
            .transition()
            .duration(ChartD3Service.focusLength)
            .style('opacity', '0');
    }
    static showFocus(svg, activeIndex) {
        ChartD3Service.hideFocus(svg);
        svg.selectAll(`.point-${activeIndex}`)
            .transition()
            .duration(ChartD3Service.focusLength)
            .attr('r', 4)
            .style('opacity', '1');
        svg.selectAll(`.scrubber-${activeIndex}`)
            .transition()
            .duration(ChartD3Service.focusLength)
            .style('opacity', '1');
    }
    static calculateLeftPad(dims, axisConfigs, data) {
        const maxYValueLength = Math.max(...data.map(d => d.y)).toString().length;
        this.yPad = maxYValueLength * 4;
        if (axisConfigs.y.showAxis || axisConfigs.y.showTicks) {
            return this.yPad + dims.margin.left;
        }
        else {
            return 0;
        }
    }
    static calculateTicks(dim) {
        return Math.max(dim / 100, 2);
    }
    static formatDataAsStacked(rawData) {
        const stackFactory = d3.stack()
            .offset(ChartD3Service.stackOffsetDiverging)
            .keys(rawData.map((_s, i) => `namey-${i}`));
        const data = ChartD3Service.seriesAsStackFormat(rawData);
        return stackFactory(data);
    }
    static seriesAsStackFormat(series) {
        const namedSeries = series.map((s, i) => {
            return {
                name: `namey-${i}`,
                coords: s
            };
        });
        const collector = [];
        namedSeries.forEach((s) => {
            s.coords.forEach((coords, i) => {
                if (!collector[i]) {
                    collector[i] = {};
                }
                collector[i].categoryValue = coords.x;
                collector[i][s.name] = coords.y;
            });
        });
        return collector;
    }
    static stackOffsetDiverging(series, seriesOrder) {
        const n = series.length;
        if (n < 1) {
            return;
        }
        const firstSeriesIndex = seriesOrder[0];
        for (let i, j = 0, dy, yp, yn, columnCount = series[firstSeriesIndex].length; j < columnCount; ++j) {
            for (yp = yn = 0, i = 0; i < n; ++i) {
                const serie = series[seriesOrder[i]];
                const dataPoint = serie[j];
                dy = dataPoint[1] - dataPoint[0];
                if (dy >= 0) {
                    dataPoint[0] = yp;
                    dataPoint[1] = yp += dy;
                }
                else if (dy < 0) {
                    dataPoint[1] = yn;
                    dataPoint[0] = yn += dy;
                }
                else {
                    dataPoint[0] = yp;
                }
            }
        }
    }
}
ChartD3Service.colors = [
    '#384b5d',
    '#6ec5ab',
    '#6D95A0',
    '#995D81',
    '#586BA4',
    '#58A4B0',
    '#D496A7',
    '#C8AD55',
    '#8EA4D2',
    '#4ECDC4',
    '#98D2EB',
    '#857C8D'
];
ChartD3Service.animationLength = 2500;
ChartD3Service.focusLength = 400;
ChartD3Service.yPad = 15;
ChartD3Service.groupPadding = 2;

class AxisEmitter {
    constructor(emitter) {
        this.emitter = emitter;
    }
    emit(axisConfig) {
        return this.emitter.emit(axisConfig);
    }
}
class AxisStrategy extends AxisEmitter {
}
class XAxisStrategy extends AxisStrategy {
    addAxis(svg, dims, data, axisConfigs, minMax, primarySeries, y) {
        return ChartD3Service.addXAxis(svg, dims, data, axisConfigs, minMax, primarySeries, y);
    }
}
class YAxisStrategy extends AxisStrategy {
    addAxis(svg, dims, data, axisConfigs, minMax) {
        return ChartD3Service.addYAxis(svg, dims, data, axisConfigs, minMax);
    }
}

class Axis {
    constructor() {
        this.dimension = 'x';
        this.show = true;
        this.ticks = true;
    }
    componentDidLoad() {
        this.setStrategy();
        const axisConfig = {
            showAxis: this.show,
            showTicks: this.ticks,
            formattingOptions: {
                locale: this.locale || 'en-US',
                language: this.language || 'en',
                currency: this.currency || 'USD'
            }
        };
        if (this.type) {
            axisConfig.labelType = this.type;
        }
        this.service.emit(axisConfig);
    }
    async init(svg, dims, data, axisConfigs, minMax, primarySeries, y) {
        return Promise.resolve(this.service.addAxis(svg, dims, data, axisConfigs, minMax, primarySeries, y));
    }
    setStrategy() {
        this.service = (this.dimension === 'x') ? new XAxisStrategy(this.anjXAxisDidLoad) : new YAxisStrategy(this.anjYAxisDidLoad);
    }
    static get is() { return "anj-axis"; }
    static get properties() { return {
        "currency": {
            "type": String,
            "attr": "currency"
        },
        "dimension": {
            "type": String,
            "attr": "dimension",
            "reflectToAttr": true
        },
        "element": {
            "elementRef": true
        },
        "init": {
            "method": true
        },
        "language": {
            "type": String,
            "attr": "language"
        },
        "locale": {
            "type": String,
            "attr": "locale"
        },
        "show": {
            "type": Boolean,
            "attr": "show"
        },
        "ticks": {
            "type": Boolean,
            "attr": "ticks"
        },
        "type": {
            "type": String,
            "attr": "type"
        }
    }; }
    static get events() { return [{
            "name": "anjYAxisDidLoad",
            "method": "anjYAxisDidLoad",
            "bubbles": true,
            "cancelable": true,
            "composed": true
        }, {
            "name": "anjXAxisDidLoad",
            "method": "anjXAxisDidLoad",
            "bubbles": true,
            "cancelable": true,
            "composed": true
        }]; }
}

class ElementStrategyFactory {
    static get(primaryType, primarySeries) {
        const data = primarySeries.map(c => c.data);
        switch (primaryType) {
            case SeriesType.Line:
                return (data.length > 1) ? new MultiLineSeries(data) : new SingleLineSeries(data);
            case SeriesType.Stacked:
                return new StackedSeries(data);
            case SeriesType.Column:
            default:
                return (data.length > 1) ? new MultiColumnSeries(data) : new SingleColumnSeries(data);
        }
    }
}
class CoreFuncs {
    constructor(data) {
        this.data = data;
    }
    coreFormatData(axisConfigs, data) {
        const evalData = data || this.data;
        return (Array.isArray(evalData[0])) ? evalData.map(d => ChartDataService.formatData(axisConfigs, d)) : ChartDataService.formatData(axisConfigs, evalData);
    }
    formatPrimaryData(axisConfigs, data) {
        return { data: this.coreFormatData(axisConfigs, data) };
    }
    flatten(data) {
        return [].concat(...data);
    }
    getTargetingData(data) {
        return data[0];
    }
    xAxisTypeOverride(axisConfigs) {
        if (axisConfigs.x.labelType === 'date' && isNaN(Number(new Date(this.data[0][0].x)))) {
            axisConfigs.x.labelType = 'string';
        }
        return axisConfigs;
    }
    getDims(chartContainer) {
        const margin = { top: 20, right: 20, bottom: 20, left: 20 };
        const { height, width } = chartContainer.getBoundingClientRect();
        return {
            height: height - margin.top - margin.bottom,
            width: width - margin.left - margin.right,
            margin
        };
    }
    getDefaultXAxisConfig() {
        return {
            showAxis: false,
            labelType: 'date',
        };
    }
    getDefaultYAxisConfig() {
        return {
            showAxis: false,
            labelType: 'number',
        };
    }
}
class SingleSeriesChartStrategy extends CoreFuncs {
    arrow() {
        return true;
    }
    getPopoverEl(svg, index) {
        return svg.select('.point-' + index).nodes()[0];
    }
    getPopoverPlacement() {
        return 'top';
    }
    getSeriesStyle() {
        return SeriesStyle.Single;
    }
    isMultiSeries() {
        return false;
    }
    getPopoverSeriesStyle() {
        return this.getSeriesStyle();
    }
}
class MultiSeriesChartStrategy extends CoreFuncs {
    arrow() {
        return false;
    }
    getPopoverEl(svg, index) {
        return svg.select('.scrubber-' + index).nodes()[0];
    }
    getPopoverPlacement() {
        return 'right';
    }
    getSeriesStyle() {
        return SeriesStyle.Multi;
    }
    isMultiSeries() {
        return true;
    }
    getPopoverSeriesStyle() {
        return this.getSeriesStyle();
    }
}
class StackedSeries extends MultiSeriesChartStrategy {
    formatPrimaryData(axisConfigs, data) {
        const baseFormat = this.coreFormatData(axisConfigs, data);
        const stackedData = ChartD3Service.formatDataAsStacked(baseFormat);
        const flattenedPointValues = stackedData.flat().flat();
        const minMax = { min: Math.min(...flattenedPointValues), max: Math.max(...flattenedPointValues) };
        return {
            data: baseFormat,
            stackedData,
            minMax
        };
    }
}
class SingleLineSeries extends SingleSeriesChartStrategy {
    xAxisTypeOverride(axisConfigs) {
        axisConfigs.x.labelType = 'date';
        return axisConfigs;
    }
}
class MultiLineSeries extends MultiSeriesChartStrategy {
    xAxisTypeOverride(axisConfigs) {
        axisConfigs.x.labelType = 'date';
        return axisConfigs;
    }
}
class SingleColumnSeries extends SingleSeriesChartStrategy {
    getPopoverSeriesStyle() {
        return SeriesStyle.Single;
    }
}
class MultiColumnSeries extends MultiSeriesChartStrategy {
    arrow() {
        return true;
    }
    getPopoverPlacement() {
        return 'top';
    }
    getPopoverSeriesStyle() {
        return SeriesStyle.Single;
    }
}

class ChartEvents {
    constructor(primarySeries, anjHover, chartService, popover) {
        this.primarySeries = primarySeries;
        this.anjHover = anjHover;
        this.chartService = chartService;
        this.popover = popover;
    }
    anjHoverHandler(svg, data, x, axisConfigs, seriesMeta) {
        const targetData = this.chartService.getTargetingData(data);
        const dataIndex = ChartD3Service.getIndexOfClosestPoint(x, targetData);
        if (this.activeIndex !== dataIndex) {
            this.setPopoverData(dataIndex, data, axisConfigs, seriesMeta);
            this.popoverTimer = window.setTimeout(() => {
                this.displayPopover(svg, this.activeIndex, true);
            }, ChartEvents.popoverDebounceLength);
        }
    }
    anjColumnHoverHandler(targetEl, data, dataIndex, axisConfigs, seriesMeta) {
        window.clearTimeout(this.popoverTimer);
        this.setPopoverData(dataIndex, [data], axisConfigs, seriesMeta);
        this.displayPopover(targetEl);
    }
    anjStackedHoverHandler(svg, data, dataIndex, axisConfigs, seriesMeta) {
        window.clearTimeout(this.popoverTimer);
        if (this.activeIndex !== dataIndex) {
            this.setPopoverData(dataIndex, data, axisConfigs, seriesMeta);
            this.displayPopover(svg, this.activeIndex, true);
        }
    }
    anjEndHoverHandler(svg) {
        switch (this.primarySeries) {
            case SeriesType.Line:
                window.clearTimeout(this.popoverTimer);
                this.hidePopover(svg);
            default:
                this.popoverTimer = window.setTimeout(() => {
                    this.hidePopover(svg);
                }, ChartEvents.popoverDebounceLength);
        }
    }
    displayPopover(svg, activeIndex, focus) {
        const popoverEl = (!isNaN(activeIndex)) ? this.chartService.getPopoverEl(svg, activeIndex) : svg;
        this.display(popoverEl);
        if (focus) {
            ChartD3Service.showFocus(svg, activeIndex);
        }
    }
    async display(target) {
        if (!target) {
            return;
        }
        const popper = await this.popover.popper();
        if (popper) {
            popper.reference = target;
            popper.scheduleUpdate();
        }
        else {
            this.popover.open(target);
        }
    }
    hidePopover(svg) {
        this.activeIndex = null;
        this.popover.close();
        ChartD3Service.hideFocus(svg);
    }
    setPopoverData(dataIndex, data, axisConfigs, seriesMeta) {
        this.activeIndex = dataIndex;
        const seriesStyle = this.chartService.getPopoverSeriesStyle();
        const activeData = { data, axisConfigs, activeIndex: this.activeIndex, seriesStyle, seriesMeta };
        this.anjHover.emit(activeData);
    }
}
ChartEvents.popoverDebounceLength = 150;

class Chart {
    constructor() {
        this.service = new SingleLineSeries();
    }
    componentWillLoad() {
        this.axisConfigs = {
            x: this.service.getDefaultXAxisConfig(),
            y: this.service.getDefaultYAxisConfig()
        };
    }
    componentDidLoad() {
        this.debounceCreation();
    }
    async onColumnsLoaded() {
        this.columns = await this.loadColumns();
        this.debounceCreation();
    }
    async onLinesLoaded() {
        this.lines = await this.loadLines();
        this.debounceCreation();
    }
    async onStackedColumnsLoaded() {
        this.stackedColumns = await this.loadStackedColumns();
        this.debounceCreation();
    }
    async chartComponentsUpdate() {
        this.popoverComponent = await this.loadPopover();
    }
    async onXAxisLoaded({ detail }) {
        this.xAxisElement = await this.loadXAxis();
        this.axisConfigs.x = Object.assign({}, this.axisConfigs.x, detail);
        this.debounceCreation();
    }
    async onYAxisLoaded({ detail }) {
        this.yAxisElement = await this.loadYAxis();
        this.axisConfigs.y = Object.assign({}, this.axisConfigs.y, detail);
        this.debounceCreation();
    }
    handleResize() {
        this.debounceCreation();
    }
    addLineEvents(svg, chartOverlay, x, data) {
        chartOverlay
            .on('mousemove', () => this.eventsService.anjHoverHandler(svg, data, x, this.axisConfigs, this.seriesMeta))
            .on('mouseout', () => this.eventsService.anjEndHoverHandler(svg));
    }
    addColumnEvents(svgBars, data) {
        svgBars
            .on('mousemove', (_d, i, bars) => this.eventsService.anjColumnHoverHandler(bars[i], data, i, this.axisConfigs, this.seriesMeta))
            .on('mouseout', () => this.eventsService.anjEndHoverHandler(svgBars));
    }
    addStackedEvents(svg, bars) {
        bars
            .on('mousemove', (_d, i) => this.eventsService.anjStackedHoverHandler(svg, this.popoverData, i, this.axisConfigs, this.seriesMeta))
            .on('mouseout', () => this.eventsService.anjEndHoverHandler(svg));
    }
    appendSVG(svg) {
        while (this.chartContainer.children.length > 1) {
            this.chartContainer.removeChild(this.chartContainer.lastChild);
        }
        this.chartContainer.appendChild(svg.node());
    }
    debounceCreation() {
        window.clearTimeout(this.creationTimer);
        this.creationTimer = window.setTimeout(async () => {
            this.chartSVG = await this.create();
        }, Chart.creationDebounceLength);
    }
    async create() {
        const dims = this.service.getDims(this.chartContainer);
        this.primarySeries = (this.stackedColumns) ? SeriesType.Stacked : (this.lines && !this.columns) ? SeriesType.Line : SeriesType.Column;
        this.secondarySeries = ((this.stackedColumns || this.columns) && this.lines) ? SeriesType.Line : undefined;
        this.primaryComponents = (this.stackedColumns) ? this.stackedColumns : (this.lines && !this.columns) ? this.lines : this.columns;
        this.secondaryComponents = ((this.stackedColumns || this.columns) && this.lines) ? this.lines : undefined;
        this.service = ElementStrategyFactory.get(this.primarySeries, this.primaryComponents);
        this.multiSeries = this.service.isMultiSeries();
        this.popover.placement = this.service.getPopoverPlacement();
        this.axisConfigs = this.service.xAxisTypeOverride(this.axisConfigs);
        this.eventsService = new ChartEvents(this.primarySeries, this.anjHover, this.service, this.popover);
        const { data, stackedData, minMax } = this.service.formatPrimaryData(this.axisConfigs);
        const flattenedData = this.service.flatten(data);
        this.popoverData = data;
        this.stackedData = stackedData;
        const svg = ChartD3Service.createSVG(dims);
        const y = (this.yAxisElement)
            ? await this.hydrateYAxis(dims, flattenedData, svg, minMax)
            : ChartD3Service.addYAxis(svg, dims, flattenedData, this.axisConfigs, minMax);
        const x = (this.xAxisElement)
            ? await this.hydrateXAxis(dims, flattenedData, svg, this.primarySeries, y)
            : ChartD3Service.addXAxis(svg, dims, flattenedData, this.axisConfigs, undefined, this.primarySeries, y);
        await this.hydratePrimaryComponents(svg, x, y, data, dims);
        await this.hydrateSecondaryComponents(svg, x, y, dims);
        this.appendSVG(svg);
        return Promise.resolve(svg);
    }
    async hydratePrimaryComponents(svg, x, y, data, dims) {
        switch (this.primarySeries) {
            case SeriesType.Column:
                const x1 = ChartD3Service.getX1(x, data, this.multiSeries);
                return this.hydrateColumns(svg, data, x, x1, y, dims);
            case SeriesType.Stacked:
                return this.hydrateStackedColumns(svg, this.stackedData, x, y, dims);
            case SeriesType.Line:
                return this.hydrateLinesPrimary(svg, data, x, y, dims);
            default:
                return Promise.resolve();
        }
    }
    async hydrateSecondaryComponents(svg, x, y, dims) {
        switch (this.secondarySeries) {
            case SeriesType.Line:
                const lineData = this.secondaryComponents.map(l => l.data);
                const formattedLineData = this.service.coreFormatData(this.axisConfigs, lineData);
                await this.hydrateLinesSecondary(svg, formattedLineData, x, y, dims);
            default:
                return Promise.resolve();
        }
    }
    async hydrateColumns(svg, data, x, x1, y, dims) {
        this.seriesMeta = [];
        this.columns.forEach(async (column, i) => {
            const bars = await column.init(svg, x, x1, y, data[i], i, dims);
            this.seriesMeta[i] = { name: column.name, netNegative: column.netNegative, color: undefined };
            this.addColumnEvents(bars, data[i]);
        });
    }
    async hydrateLinesPrimary(svg, data, x, y, dims) {
        this.seriesMeta = [];
        this.lines.forEach(async (line, i) => {
            const color = await line.init(svg, x, y, data[i], i, dims, this.service.isMultiSeries());
            this.seriesMeta[i] = { name: line.name, color, netNegative: line.netNegative };
        });
        const chartOverlay = ChartD3Service.addOverlay(svg, dims);
        this.addLineEvents(svg, chartOverlay, x, data);
        ChartD3Service.addRevealAnimation(svg, dims, this.axisConfigs);
    }
    async hydrateLinesSecondary(svg, data, x, y, dims) {
        this.lines.forEach(async (line, i) => {
            this.popoverData.push(data[i]);
            const color = await line.init(svg, x, y, data[i], i, dims, this.service.isMultiSeries(), this.primaryComponents.length);
            const popoverIndex = i + this.seriesMeta.length;
            this.seriesMeta[popoverIndex] = { name: line.name, color, netNegative: line.netNegative };
        });
    }
    async hydrateStackedColumns(svg, data, x, y, dims) {
        this.seriesMeta = [];
        this.stackedColumns.forEach(async (column, i) => {
            const { bars, color } = await column.init(svg, x, y, data[i], i, dims);
            this.seriesMeta[i] = { name: column.name, netNegative: column.netNegative, color };
            this.addStackedEvents(svg, bars);
        });
    }
    hydrateXAxis(dims, data, svg, primarySeries, y) {
        return this.xAxisElement.init(svg, dims, data, this.axisConfigs, undefined, primarySeries, y);
    }
    hydrateYAxis(dims, data, svg, minMax) {
        return this.yAxisElement.init(svg, dims, data, this.axisConfigs, minMax, undefined);
    }
    loadColumns() {
        return Promise.all(Array.from(this.chartContainer.querySelectorAll('anj-column-series')).map(col => col.componentOnReady()));
    }
    loadLines() {
        return Promise.all(Array.from(this.chartContainer.querySelectorAll('anj-line-series')).map(col => col.componentOnReady()));
    }
    loadStackedColumns() {
        return Promise.all(Array.from(this.chartContainer.querySelectorAll('anj-stacked-column-series')).map(col => col.componentOnReady()));
    }
    loadPopover() {
        return Promise.resolve(this.chartContainer.querySelector('anj-popover-content-single, anj-popover-content-multi').componentOnReady());
    }
    loadXAxis() {
        return Promise.resolve(this.chartContainer.querySelector('anj-axis[dimension="x"]').componentOnReady());
    }
    loadYAxis() {
        return Promise.resolve(this.chartContainer.querySelector('anj-axis[dimension="y"]').componentOnReady());
    }
    hostData() {
        return {
            class: {
                'anj-rm-pop-arrow': !this.service.arrow()
            }
        };
    }
    render() {
        return [
            h("anj-overlay", { class: "popover", placement: "top", offset: "0, 15", ref: r => this.popover = r }, (this.popoverComponent) ? [h("anj-arrow", null), h("slot", null)] : undefined)
        ];
    }
    static get is() { return "anj-chart"; }
    static get properties() { return {
        "chartContainer": {
            "elementRef": true
        },
        "multiSeries": {
            "state": true
        },
        "popoverComponent": {
            "state": true
        }
    }; }
    static get events() { return [{
            "name": "anjHover",
            "method": "anjHover",
            "bubbles": true,
            "cancelable": true,
            "composed": true
        }]; }
    static get listeners() { return [{
            "name": "anjColumnDidLoad",
            "method": "onColumnsLoaded"
        }, {
            "name": "anjLineDidLoad",
            "method": "onLinesLoaded"
        }, {
            "name": "anjStackedColumnDidLoad",
            "method": "onStackedColumnsLoaded"
        }, {
            "name": "anjPopoverDidLoad",
            "method": "chartComponentsUpdate"
        }, {
            "name": "anjXAxisDidLoad",
            "method": "onXAxisLoaded"
        }, {
            "name": "anjYAxisDidLoad",
            "method": "onYAxisLoaded"
        }, {
            "name": "window:resize",
            "method": "handleResize",
            "passive": true
        }]; }
    static get style() { return "anj-chart{display:block;height:100%;position:relative;width:100%}anj-chart .overlay{cursor:default;fill:none;pointer-events:all;position:relative;z-index:3}anj-chart svg .area{opacity:.1}anj-chart svg .axis{font-size:.5rem}anj-chart svg .axis line,anj-chart svg .axis path{fill:none;stroke:var(--anj-border);shape-rendering:crispEdges}anj-chart svg .axis.y path{stroke:transparent}anj-chart svg .tick text,anj-chart svg text{fill:var(--anj-neutral-light)}anj-chart svg .line{fill:none;stroke-width:2px;-webkit-transition:opacity .5s ease,stroke-width .5s ease;transition:opacity .5s ease,stroke-width .5s ease}anj-chart svg .overlay{fill:none;pointer-events:all;position:relative;z-index:3}anj-chart svg .point{fill:var(--anj-light);stroke-width:1.5px}anj-chart.anj-rm-pop-arrow anj-arrow{display:none}anj-chart .line,anj-chart .popover{pointer-events:none}anj-chart .series-container .bar{opacity:1;-webkit-transition:opacity .5s;transition:opacity .5s}"; }
}
Chart.creationDebounceLength = 100;

class ColumnSeries {
    dataHandler() {
        this.anjColumnDidLoad.emit();
    }
    componentDidLoad() {
        this.anjColumnDidLoad.emit();
    }
    async init(svg, x, x1, y, data, index, dims) {
        this.color = this.color || ChartD3Service.colors[index || 0];
        const bars = ChartD3Service.addColumn(svg, x, x1, y, data, this.color, dims, index);
        return Promise.resolve(bars);
    }
    static get is() { return "anj-column-series"; }
    static get properties() { return {
        "color": {
            "type": String,
            "attr": "color",
            "mutable": true
        },
        "data": {
            "type": "Any",
            "attr": "data",
            "watchCallbacks": ["dataHandler"]
        },
        "element": {
            "elementRef": true
        },
        "init": {
            "method": true
        },
        "name": {
            "type": String,
            "attr": "name"
        },
        "netNegative": {
            "type": Boolean,
            "attr": "net-negative"
        }
    }; }
    static get events() { return [{
            "name": "anjColumnDidLoad",
            "method": "anjColumnDidLoad",
            "bubbles": true,
            "cancelable": true,
            "composed": true
        }]; }
}

class LineSeries {
    dataHandler() {
        this.anjLineDidLoad.emit();
    }
    componentDidLoad() {
        this.anjLineDidLoad.emit();
    }
    async init(svg, x, y, data, index, dims, multiSeries, secondaryIndex) {
        this.color = this.color || ChartD3Service.colors[secondaryIndex || index || 0];
        const line = ChartD3Service.addLine(svg, x, y, data, this.color);
        const allPositive = (data) ? Math.min(...data.map(d => Number(d.y))) >= 0 : true;
        if (multiSeries) {
            ChartD3Service.addScrubbers(svg, x, data, dims, index);
        }
        else if (allPositive) {
            ChartD3Service.addArea(svg, line, y, data, this.color);
        }
        ChartD3Service.addPoints(svg, x, y, data, this.color, index);
        return Promise.resolve(this.color);
    }
    static get is() { return "anj-line-series"; }
    static get properties() { return {
        "color": {
            "type": String,
            "attr": "color",
            "mutable": true
        },
        "data": {
            "type": "Any",
            "attr": "data",
            "watchCallbacks": ["dataHandler"]
        },
        "element": {
            "elementRef": true
        },
        "init": {
            "method": true
        },
        "name": {
            "type": String,
            "attr": "name"
        },
        "netNegative": {
            "type": Boolean,
            "attr": "net-negative"
        }
    }; }
    static get events() { return [{
            "name": "anjLineDidLoad",
            "method": "anjLineDidLoad",
            "bubbles": true,
            "cancelable": true,
            "composed": true
        }]; }
}

class StackedColumnSeries {
    dataHandler() {
        this.anjStackedColumnDidLoad.emit();
    }
    componentDidLoad() {
        this.anjStackedColumnDidLoad.emit();
    }
    async init(svg, x, y, data, index, dims) {
        this.color = this.color || ChartD3Service.colors[index || 0];
        const bars = ChartD3Service.addStackedColumn(svg, x, y, data, this.color);
        ChartD3Service.addStackedScrubbers(svg, x, data, dims, index);
        return Promise.resolve({ bars, color: this.color });
    }
    static get is() { return "anj-stacked-column-series"; }
    static get properties() { return {
        "color": {
            "type": String,
            "attr": "color",
            "mutable": true
        },
        "data": {
            "type": "Any",
            "attr": "data",
            "watchCallbacks": ["dataHandler"]
        },
        "element": {
            "elementRef": true
        },
        "init": {
            "method": true
        },
        "name": {
            "type": String,
            "attr": "name"
        },
        "netNegative": {
            "type": Boolean,
            "attr": "net-negative"
        }
    }; }
    static get events() { return [{
            "name": "anjStackedColumnDidLoad",
            "method": "anjStackedColumnDidLoad",
            "bubbles": true,
            "cancelable": true,
            "composed": true
        }]; }
}

export { Axis as AnjAxis, Chart as AnjChart, ColumnSeries as AnjColumnSeries, LineSeries as AnjLineSeries, StackedColumnSeries as AnjStackedColumnSeries };
